import csv
import os
from queue import Queue
import shutil
import sys
import subprocess
import threading

# Directories and files
root_dir = '/home/mx/WALI/stuff/package_stuff/'
package_dir = root_dir + 'packages/'
log_dir = root_dir + 'logs/'
dlog_dir = log_dir + 'deps/'
libs_dir = root_dir + 'libs/'
deps_dir = root_dir + 'deps/'
aports_dir = '/home/mx/aports/main/'
package_file = root_dir + 'packages.txt' # default
whitelist_file = root_dir + 'whitelist'
csuccess_file = os.path.join(package_dir, 'csuccess.txt')
msuccess_file = os.path.join(package_dir, 'msuccess.txt')

# commands for configure/make
sysroot_dir = '/home/mx/WALI/WALI/wali-musl/sysroot'
cflag_warning_ign = "-Wno-implicit-function-declaration -Wno-int-conversion -Wno-incompatible-function-pointer-types"
wali_cflags = f"-O0 --target=wasm32-wasi-threads -pthread {cflag_warning_ign} --sysroot={sysroot_dir} -L{sysroot_dir}/lib -matomics -mbulk-memory -mmutable-globals -msign-ext"
wali_lflags = "-Wl,--shared-memory -Wl,--export-memory -Wl,--max-memory=67108864 -lpthread"
conf_command = ("./configure "
                "CC=/home/mx/WALI/WALI/llvm-project/build/bin/clang "
                "CC_FOR_BUILD=/home/mx/WALI/WALI/llvm-project/build/bin/clang "
                f"CFLAGS=\"{wali_cflags}\" "
                "LD=wasm-ld "
                f"LDFLAGS=\"{wali_cflags} {wali_lflags}\" "
                "LIBS=\"-lpthread\" "
                "AR=llvm-ar RANLIB=llvm-ranlib "
                "cross_compiling=yes "
                "&& make")

make_command = f"make CC=/home/mx/WALI/WALI/llvm-project/build/bin/clang CC_FOR_BUILD=/home/mx/WALI/WALI/llvm-project/build/bin/clang CFLAGS=\"{wali_cflags}\" LD=wasm-ld LDFLAGS=\"{wali_cflags} {wali_lflags}\" LIBS=\"-lpthread\" AR=llvm-ar RANLIB=llvm-ranlib cross_compiling=yes"
print(make_command)
exit(0)
# Create the packages and libs directories if they don't exist
os.makedirs(package_dir, exist_ok=True)
os.makedirs(libs_dir, exist_ok=True)
os.makedirs(deps_dir, exist_ok=True)
os.makedirs(dlog_dir, exist_ok=True)

# Threading setup
NUM_THREADS = 32
queue = Queue()
lock = threading.Lock()

# debug
dbg = "-d" in sys.argv
def dbg_print(s):
    if dbg:
        with lock:
            print(s, flush=True) # thread safe printing

# usage
# TODO: use it somewhere
def usage():
    print(f'{sys.argv[0]} - a program that does things related to WALI packages\n')
    print('\t-d\t\tPrints information for debugging.\n')
    print('\t-D\t\tBuild and save dependency trees for packages.\n')
    print('\t-L\t\tBuild static libraries given a list of libs.\n')
    print('\t-t\t\tBuild one package for testing purposes, logs will be printed to stdout.\n')
    exit(0)

# Tracking success and failure
total_packages = 0
configure_success = 0
make_success = 0
dependency_failures = 0

# declare the list of packages and all available packages
# which is the default package file
packages = []
with open(whitelist_file, 'r') as f:
    whitelist = [p for p in f.readlines()]
with open(package_file, 'r') as f:
    available_packages = [p for p in f.readlines()]

# read package file from user and setup related things
def get_packages():
    global total_packages, package_file, packages
    # Read the package list
    package_file = str(input('Package file:')).split('/')[-1]
    with open(package_file, 'r') as f:
        packages = [line.strip() for line in f if line.strip()]
    pnum = int(input("How many packages(-1 for all): "))
    tnum = len(packages)
    dbg_print(f'compiling {pnum}/{tnum} packages...\n')
    if pnum < 0 or pnum > tnum:
        total_packages = tnum
    else:
        total_packages = pnum

# record the number of parents each package has
num_parents = {}

def log_message(log_file, message):
    if log_file == sys.stdout:
        with lock:
            print(message, flush=True)
        return
    with open(log_file, 'a') as log:
        log.write(message + '\n')

# run command, returns stdout, stderr, success
def run_command(command, cwd=None):
    try:
        result = subprocess.run(command, shell=True, cwd=cwd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.stdout.decode('utf-8'), result.stderr.decode('utf-8'), result.returncode == 0
    except subprocess.CalledProcessError as e:
        return e.stdout.decode('utf-8'), e.stderr.decode('utf-8'), False

# @param: lf: log file; package: name of the package trying to resolve
def resolve_dependencies(package, lf): # TODO: use trees instead
    # new plan: 
    # 1. get deps (from tree possibly (add second representation))
    # 2. check if each dep is done (in a file? maintain a list?)
    # 3. if not, process it

    deps_dir = os.path.join(package_dir, package, 'dependencies')
    os.makedirs(deps_dir, exist_ok=True)

    stdout, stderr, success = run_command(f"apk info -R {package}")
    log_message(lf, 'Error running apt-cache depends.')
    if not success:
        return False

    dependencies = []
    for line in stdout.splitlines():
        if 'Depends:' in line:
            dep = line.split('Depends:')[1].strip()
            dependencies.append(dep)

    all_deps_compiled = True
    for dep in dependencies:
        dep_dir = os.path.join(deps_dir, dep)
        os.makedirs(dep_dir, exist_ok=True)
        dep_stdout, dep_stderr, dep_success = run_command(f"apt source {dep}", dep_dir)
        
        if not dep_success:
            all_deps_compiled = False
            continue
        
        dep_things = os.listdir(dep_dir)
        dep_ap_dir = ''
        for f in dep_things:
            cur_f = os.path.join(dep_dir, f)
            if os.path.isdir(cur_f):
                dep_ap_dir = cur_f
                break

        if not dep_ap_dir:
            all_deps_compiled = False
            continue
        
        stdout, stderr, success, _ = configure_and_build_package(dep_ap_dir)
        log_message(lf, f'[{dep}]:\nstdout:{stdout}\nstderr:{stderr}\n')
        if not success:
            all_deps_compiled = False

    return all_deps_compiled

def configure_and_build_package(package_dir, buildlib=False, libdir='', libname=''):
    configure_script = os.path.join(package_dir, 'configure')
    static = f'{libname}.a'

    makefile = os.path.join(package_dir, 'Makefile')
    if os.path.exists(configure_script):
        command = conf_command
    elif os.path.exists(makefile):
        command = make_command
        if buildlib: 
            # ignore confable packages for now 
            # since (i think) usually they don't have .a option 
            # (Makefile is way too complicated to read)
            command += f' {static}'
    else:
        return '', f'Error: not configurable/makeable.\nDirectory list:\n{os.listdir(package_dir)}\n', False, False

    stdout, stderr, success = run_command(command, cwd=package_dir)


    if buildlib and os.path.exists(os.path.join(package_dir, static)):
        shutil.move(os.path.join(package_dir, static), libdir)

    return stdout, stderr, success, os.path.exists(configure_script)

def cleanup(p_dir):
    if '-t' in sys.argv:
        return
    try:
        shutil.rmtree(p_dir)
        print('.', end='', flush=True)
    except OSError as e:
        print("Error: %s - %s." % (e.filename, e.strerror))

# return true if package not found in aport thingy
def is_subpackage(package):
    return (package not in available_packages)

# pre - is subpackage (is_subpackage returned false)
# TODO: error check two calls to run_command
def find_origin(package, logfile):
    # get package dir
    pdir = package_dir + package
    the_apk_file = ''

    # apk fetch
    run_command(f'apk fetch {package}', cwd=pdir)

    # find .apk file, TODO: implement find_in_dir function
    things_in_dir = os.listdir(pdir)
    for thing in things_in_dir:
        if thing.endswith('.apk'):
            the_apk_file = thing
    
    if not the_apk_file:
        log_message(logfile, 'Failed to apk fetch.\n')
        return ''

    # tar -xvf
    run_command(f'tar -xvf {the_apk_file}', cwd=pdir)

    # read .PKGINFO
    with open(os.path.join(pdir, '.PKGINFO'), 'r') as f:
        # get origin
        lines = f.read()
        for line in lines.split('\n'):
            if 'origin' in line:
                return line.split(' ')[-1]
    
    print(logfile) # DEBUG
            
    log_message(logfile, f'Origin doesn\'t exist.\n.PKGINFO:{lines}') # shouldn't happen
    return ''

def process_package(package, outstd=False):
    global configure_success, make_success, dependency_failures

    p_dir = os.path.join(package_dir, package)
    os.makedirs(p_dir, exist_ok=True)
    log_file = sys.stdout if outstd else os.path.join(log_dir, package + '.log')

    # if building libraries, save old dir and package for future use
    if '-L' in sys.argv:
        old_package = package
        old_pdir = p_dir
        old_logfile = log_file
    
    # package is a subpackage (or doesn't exist, shouldn't happen but just in case?)
    if is_subpackage(package):
        # print('DDDDDDDDDDDDDDDEBUGGGGGGGGGGGGGGG', flush=True) # DEBUG
        # print(f'sub package:{package}', flush=True) # DEBUG
        package = find_origin(package, log_file)
        # print(f'parent package:{package}\nDDDDDDDDDDDDDDDEBUGGGGGGGGGGGGGGG', flush=True) # DEBUG
        if not package:
            log_message(log_file, 'Package does not exist. \n')
            cleanup(p_dir)
            return ''

    aport_dir = os.path.join(aports_dir, package)
    p_dir = os.path.join(package_dir, package)
    os.makedirs(p_dir, exist_ok=True)
    # print(os.listdir(package_dir), flush=True) # DEBUG
    # print(f'pdir: {p_dir}', flush=True) # DEBUG
    log_file = sys.stdout if outstd else os.path.join(log_dir, package + '.log')

    # Get the source package
    # for Alpine: set srcdest for each package before cloning source
        # maybe lock until source get
        # or run in subshell which have different env vars
    dbg_print(f'[{package}] getting source package...')
    stdout, stderr, success = run_command(f'export SRCDEST="{p_dir}" && abuild fetch', cwd=aport_dir)
    # print(stdout, flush=True) # DEBUG
    # with lock:
        # stdout, se, su = run_command(f'echo $SRCDEST') # DEBUG
        # print(f'srcdest: {stdout}', flush=True) # DEBUG
        # stdout, stderr, success = run_command(f"abuild fetch {package}", aport_dir)
        # print(os.listdir(p_dir), flush=True) # DEBUG

    # extract package from tar
    the_tar = ''
    # print(os.listdir(package_dir), flush=True) # DEBUG
    # print(f'package: {package}', flush=True) # DEBUG
    # print(os.listdir(p_dir), flush=True) # DEBUG
    for f in os.listdir(p_dir):
        if '.tar.' in f:
            the_tar = f
            break
    if not the_tar:
        log_message(log_file, f"Package tar not found. {os.listdir(p_dir)}")
        cleanup(p_dir)
        return
    
    stdout, stderr, success = run_command(f"tar -xf {the_tar}", p_dir)

    if not success:
        log_message(log_file, stderr)
        cleanup(p_dir)
        return

    # set directory of actual package from source
    things = os.listdir(p_dir)
    ap_dir = ''
    for f in things:
        cur_f = os.path.join(p_dir, f)
        if os.path.isdir(cur_f) and f != 'usr':
            ap_dir = cur_f
            break

    if not ap_dir:
        log_message(log_file, f"Dir not found. {os.listdir(p_dir)}")
        cleanup(p_dir)
        return

    # Resolve dependencies, TODO: fix this, maybe through tree search thing
    # if not resolve_dependencies(package, log_file):
    #     with lock:
    #         dependency_failures += 1
    #     log_message(log_file, "Failed to resolve dependencies.")
    #     cleanup(p_dir)
    #     return

    # Configure the package

    
    if '-L' not in sys.argv:
        stdout, stderr, success, corm = configure_and_build_package(ap_dir)
    else:
        stdout, stderr, success, corm = configure_and_build_package(ap_dir, True, old_pdir, old_package)
        

    if success:
        print('.', end='')
        with lock:
            if corm:
                configure_success += 1
                with open(csuccess_file, 'a') as f:
                    f.write(package + '\n')
            else:
                make_success += 1
                with open(msuccess_file, 'a') as g:
                    g.write(package + '\n')
            # Move successful library packages to libs directory
            if package.startswith("lib"):
                lib_dest_dir = os.path.join(libs_dir, package)
                shutil.move(ap_dir, lib_dest_dir)
    else:
        log_message(log_file, stderr)
        cleanup(p_dir)
        return

# thread routine for building packages
def thread_build():
    while True:
        package = queue.get()
        if package is None:
            break

        # avoid rebuild (currently for libraries)
        if package in whitelist:
            continue
        
        log_file = os.path.join(log_dir, package + '.log')
        # Create a Timer to enforce a 120-second timeout for each package
        timer = threading.Timer(120, lambda: log_message(log_file, 'Time out.'))
        timer.start()

        process_package(package)
        timer.cancel()
        queue.task_done()

######################## build ^ ############## dep v #########################

def add_parent(child, parent):
    if child not in num_parents.keys():
        num_parents[child] = []
    elif (parent not in num_parents[child]) and (parent != ''):
        # num_parents[d] += depth + 1 # all ancestors of this package
        num_parents[child].append(parent)
        for elem in num_parents[parent]:
            num_parents[child].append(parent)

def process_dep(package, pre, parent, depth):

    log_file = os.path.join(dlog_dir, package + '.log')

    # TODO: setup numdep file

    # check if in dict, add if not
    with lock:
        if package not in num_parents.keys():
            num_parents[package] = []

    # get deps
    stdout, stderr, success = run_command(f'apk info -R {package}')
    if not success:
        log_message(log_file, f'apk info -R failed: {stderr}')
        return f'\n{pre}{package}!'

    # split to list and strip
    deps = [p.strip('!') for p in stdout.split('\n')[1:-2]]
    dbg_print(deps)
    if not deps or deps[0] == '':
        log_message(log_file, f'{package} reached end of dep tree.\n')
        return ''

    # find subdeps recursively
    res = ''
    for d in deps:
        if not d:
            continue
        # if package name with =[version], only take the front part
        elif '=' in d:
            d = d.split('=')[0]
            res += f'\n{pre}⮡ {d}'
        elif '<' in d:
            # virtual package, skip for now
            res += '<>'
        else:
            res += process_dep(d, f'{pre}| ', package, depth + 1)
            with lock:
                add_parent(d, parent)

    return res


# thread routine for getting dependency trees
def thread_dep():
    while True:
        package = queue.get()
        if package is None:
            break
        
        log_file = os.path.join(log_dir, package + '.log')
        tree_file = os.path.join(deps_dir, package + '.tree')
        res = package

        # Create a Timer to enforce a 120-second timeout for each package
        timer = threading.Timer(180, lambda: log_message(log_file, 'Time out.'))
        timer.start()

        res += process_dep(package, '', '', 0)
        log_message(tree_file, res)
        
        timer.cancel()  # Cancel the timer if the package is processed successfully within 2 mins
        if res != package:
            print('.', end='', flush=True)
        queue.task_done()

def finalize():
    # Print success rates
    print(f"Total packages: {total_packages}")
    print(f"Configure success: {configure_success} ({(configure_success/total_packages)*100:.2f}%)")
    print(f"Make success: {make_success} ({(make_success/total_packages)*100:.2f}%)")
    print(f"Dependency failures: {dependency_failures}")

    # put dep results in csv file
    # TODO: put log_collection here too, and into functions
    dep_freq = os.path.join(deps_dir, 'dep_freq.csv')
    dep_freq_res = []
    for pname in num_parents.keys():
        dep_freq_res.append([pname, len(num_parents[pname]), num_parents[pname]])

    with open(dep_freq, 'w', newline='', encoding='utf-8') as csvfile:
        csvwriter = csv.writer(csvfile)
        # Write the header
        csvwriter.writerow(['Package', 'Number of Parents'])
        # Write the log contents
        csvwriter.writerows(dep_freq_res)

# Create threads
threads = []

try:
    # one package, for testing
    if '-t' in sys.argv:
        test_package = str(input('package: '))
        process_package(test_package, True)
        exit(0)
        
    get_packages()
    for i in range(NUM_THREADS):
        t = threading.Thread(target=(thread_build if '-D' not in sys.argv else thread_dep))
        t.start()
        threads.append(t)

    # Enqueue packages
    for i in range(total_packages):
        queue.put(packages[i])

    # Block until all tasks are done
    queue.join()

    # Stop workers
    for i in range(NUM_THREADS):
        queue.put(None)
    for t in threads:
        t.join(120)
except KeyboardInterrupt:
    # Print success rates
    with lock:
        finalize()

# Print success rates
finalize()
