import os
import sys
import threading

package_list = open(sys.argv[1], 'r').read().split("\n")

pnum = -1
if len(sys.argv) > 2:
    pnum = int(sys.argv[2])

rootdir = os.getcwd()
package_dir = '/home/mx/WALI/stuff/package_stuff/packages'
log_dir = '/home/mx/WALI/stuff/package_stuff/logs'
os.chdir(package_dir)

# commands for configure/make
sysroot_dir = 'SYSROOT_DIR=/home/mx/WALI/WALI/wali-musl/sysroot'
cflag_warning_ign= "cflag_warning_ign=\"-Wno-implicit-function-declaration -Wno-int-conversion -Wno-incompatible-function-pointer-types\""
wali_cflags= "wali_cflags=\"-O0 --target=wasm32-wasi-threads -pthread $cflag_warning_ign --sysroot=$SYSROOT_DIR -L$SYSROOT_DIR/lib -matomics -mbulk-memory -mmutable-globals -msign-ext\""
wali_lflags= "wali_lflags=\"-Wl,--shared-memory -Wl,--export-memory -Wl,--max-memory=67108864 -lpthread\""
conf_command = '''./configure \
  CC=/home/mx/WALI/WALI/llvm-project/build/bin/clang \\
  CC_FOR_BUILD=/home/mx/WALI/WALI/llvm-project/build/bin/clang \\
  CFLAGS="$wali_cflags\" \\
  LD=wasm-ld \\
  LDFLAGS="$wali_cflags $wali_lflags" \\
  LIBS="-lpthread" \\
  AR=llvm-ar RANLIB=llvm-ranlib \\
  cross_compiling=yes \\'''

def log(p, s):
    with open(f'{log_dir}/{p}.log', 'w') as f:
        f.write(f"{p}: {s}\n")

# thread routine
# input: start position of package
# return: (ms, cs) = success from make, success from configure
# this function tries to compile 1000 packages starting from number start_num
# and returns number of successes
def compile_packages(start_num):
    c = 0
    ms = 0
    cs = 0
    for i in range(start_num, start_num+1000):
        pkg_name = package_list[i]
        c += 1

        # create dir and move to dir
        print(f'package: {pkg_name}')
        if pkg_name in os.listdir():
            os.system(f'rm -rf {pkg_name}')
        print(os.listdir())
        os.system(f'mkdir {pkg_name}')
        os.chdir(f'{pkg_name}')

        # apt source and move to dir
        try:
            print(f"apt source {pkg_name}")
            os.system(f"apt source {pkg_name}")
            print('\n\n\n\n')
            print(os.listdir())
            print("rm -f *.tar.gz *.tar.xz *.dsc *.asc")
            os.system("rm -f *.tar.gz *.tar.xz *.dsc *.asc")
            print(os.listdir())
            print('\n\n\n\n')
            print(f"cd {os.listdir()[0]}")
            os.chdir(os.listdir()[0])

            # check make or configure
            dirs = os.listdir()
            print(dirs)
            if "configure" in dirs:
                # setup options and flags and stuff then ./configure (or conf_commands.h)
                with open("conf_command.sh", 'w') as f:
                    f.write(f'''#!/bin/bash\n\n{sysroot_dir}\n{cflag_warning_ign}\n{wali_cflags}\n{wali_lflags}\n\n{conf_command}''')

                print(os.listdir())
                print("chmod +x conf_command.sh")
                os.system("chmod +x conf_command.sh")
                print("./conf_command.sh")
                if os.system("./conf_command.sh") != 0:
                    os.chdir(package_dir)
                    os.system(f"rm -rf {pkg_name}")
                    continue

                if os.system("make") != 0:
                    os.chdir(package_dir)
                    os.system(f"rm -rf {pkg_name}")
                    continue
                cs += 1

            elif "Makefile" in dirs:
                # setup options and flags and stuff then make
                if os.system('''make CC=/home/mx/WALI/WALI/llvm-project/build/bin/clang \\
                            CC_FOR_BUILD=/home/mx/WALI/WALI/llvm-project/build/bin/clang \\
                            CFLAGS="$wali_cflags\" \\
                            LD=wasm-ld \\
                            LDFLAGS="$wali_cflags $wali_lflags" \\
                            LIBS="-lpthread" \\
                            AR=llvm-ar RANLIB=llvm-ranlib \\
                            cross_compiling=yes''') != 0:
                    os.chdir(package_dir)
                    os.system(f"rm -rf {pkg_name}")
                    continue
                ms += 1
        except KeyboardInterrupt:
            print(f'{ms+cs}/{c}[{(ms+cs)/c * 100}%].\n')
            print(f'make: {ms}\configure: {cs}')
        except Exception as e:
            print(f"error: {str(e)}\n\n")
            log(pkg_name, str(e))
            os.chdir(package_dir)
            os.system(f"rm -rf {pkg_name}")


        os.chdir(package_dir)
        os.system(f"rm -rf {pkg_name}")

    return (ms, cs)

# threading
count = len(package_list)
ts = [threading.Thread(target=compile_packages, args=(t,)) for t in range(0, count-1000, 1000)]
success = [t.start() for t in ts]

mssum = sum([ms for (ms,cs) in success])
cssum = sum([cs for (ms,cs) in success])
ssum = mssum + cssum

print(f"Done. Success rate: {ssum}/{count}[{ssum/count * 100}%].\n")
print(f'success from make: {mssum}\tsuccess from configure: {cssum}')
