# deptree - builds the dependency tree for a package or a list of packages

# This program uses apt depends to find dependencies for a package,
# then recursively find dependencies for the dependencies until the last level.
# This program also uses threading to speed up the process. 

# Both the full dep tree of each package and how many times this package
# was depended on will be recorded in log_dir. The maximum depth of the
# dependency tree will also be recorded. 

# Example dep tree:
# p1
# |-p11
# |-p12
# | |-p121
# |-p13



import os
from queue import Queue
import shutil
import sys
import subprocess
import threading

# debug
debug = '-d' in sys.argv
def dbg_print(msg):
    if debug:
        print(msg, flush=true)

# dictionary for each package (only one in this test)
ps = {}

# 1. get children
# 2. pass in parent
# 3. each child add parent's parent list to their list

# adds a parent to child's list of parents
# pre: child in ps
def add_parent(child, parent):
    if parent in ps.keys:
        for grandparent in ps[parent]:
            ps[child].append(grandparent)

# # init
# debs_dir = '/home/mx/WALI/stuff/package_stuff/debs'
# log_dir = '/home/mx/WALI/stuff/package_stuff/logs'
# os.makedirs(debs_dir, exist_ok=True)
# package_file = str(input('Package file:')).split('/')[-1]

def process_dep(package, pre, parent):

    #log_file = os.path.join(dlog_dir, package + '.log')

    # TODO: setup numdep file

    # get deps
    so = subprocess.check_output(f'apk info -R {package}', shell=True)

    # split to list and strip
    # deps = [p.strip('!') for p in so.split('\n')[1:]]
    deps = [p for p in so.decode().split('\n')[1:]]

    print(deps)
    if not deps:
        # log_message(log_file, f'{package} reached end of dep tree.\n')
        return ''

    # find subdeps recursively
    #⤇⇒↳↪⇰➜
    res = ''
    for d in deps:
        res += f'\n{pre}⮡ {d}'
        if d == parent: # package depends on each other
            res += '='
        elif '<' in d: # virtual package, skip for now
            res += '<>'
        else:
            res += process_dep(d, f'{pre}| ', package)

    return res

p = str(input("package: "))
print(process_dep(p, '', ''))
print()