import os
from queue import Queue
import shutil
import subprocess
import threading
import time

# Directories and files
package_dir = '/home/mx/WALI/stuff/package_stuff/packages'
log_dir = '/home/mx/WALI/stuff/package_stuff/logs'
package_file = 'packages.txt'
csuccess_file = os.path.join(package_dir, 'csuccess.txt')
msuccess_file = os.path.join(package_dir, 'msuccess.txt')

# commands for configure/make
sysroot_dir = '/home/mx/WALI/WALI/wali-musl/sysroot'
cflag_warning_ign = "-Wno-implicit-function-declaration -Wno-int-conversion -Wno-incompatible-function-pointer-types"
wali_cflags = f"-O0 --target=wasm32-wasi-threads -pthread {cflag_warning_ign} --sysroot={sysroot_dir} -L{sysroot_dir}/lib -matomics -mbulk-memory -mmutable-globals -msign-ext"
wali_lflags = "-Wl,--shared-memory -Wl,--export-memory -Wl,--max-memory=67108864 -lpthread"
conf_command = ("./configure "
                "CC=/home/mx/WALI/WALI/llvm-project/build/bin/clang "
                "CC_FOR_BUILD=/home/mx/WALI/WALI/llvm-project/build/bin/clang "
                f"CFLAGS=\"{wali_cflags}\" "
                "LD=wasm-ld "
                f"LDFLAGS=\"{wali_cflags} {wali_lflags}\" "
                "LIBS=\"-lpthread\" "
                "AR=llvm-ar RANLIB=llvm-ranlib "
                "cross_compiling=yes "
                "&& make")

make_command = f"make CC=/home/mx/WALI/WALI/llvm-project/build/bin/clang CC_FOR_BUILD=/home/mx/WALI/WALI/llvm-project/build/bin/clang CFLAGS=\"{wali_cflags}\" LD=wasm-ld LDFLAGS=\"{wali_cflags} {wali_lflags}\" LIBS=\"-lpthread\" AR=llvm-ar RANLIB=llvm-ranlib cross_compiling=yes"

# Create the packages directory if it doesn't exist
os.makedirs(package_dir, exist_ok=True)

# Read the package list
with open(package_file, 'r') as f:
    packages = [line.strip() for line in f if line.strip()]

# Threading setup
NUM_THREADS = 32
queue = Queue()
lock = threading.Lock()

# Tracking success and failure
configure_success = 0
make_success = 0

pnum = int(input("How many packages(-1 for all): "))
tnum = len(packages)
total_packages = 0
if pnum < 0 or pnum > tnum:
    total_packages = tnum
else:
    total_packages = pnum

def run_command(command, cwd=None):
    try:
        result = subprocess.run(command, shell=True, cwd=cwd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.stdout.decode('utf-8'), result.stderr.decode('utf-8'), result.returncode == 0
    except subprocess.CalledProcessError as e:
        return e.stdout.decode('utf-8'), e.stderr.decode('utf-8'), False

#def resolve_dependencies(package_dir):
#    get dependencies, resolve each of them

def configure_and_build_package(package_dir):
    configure_script = os.path.join(package_dir, 'configure')
    if os.path.exists(configure_script):
        command = conf_command
    else:
        command = make_command

    stdout, stderr, success = run_command(command, cwd=package_dir)
    return stdout, stderr, success, os.path.exists(configure_script)

def cleanup(p_dir):
    try:
        shutil.rmtree(p_dir)
        print('.', end='')
    except OSError as e:
        print("Error: %s - %s." % (e.filename, e.strerror))

def process_package(package):
    global configure_success, make_success

    p_dir = os.path.join(package_dir, package)
    os.makedirs(p_dir, exist_ok=True)
    os.sleep(1) # prevent race condition in a scuffed way, will change later
    log_file = os.path.join(log_dir, package + '.log')

    # Get the source package
    stdout, stderr, success = run_command(f"apt source {package}", p_dir)

    if not success:
        with open(log_file, 'w') as log:
            log.write(stderr)
        cleanup(p_dir)
        return

    # set directory of actual package from apt source
    things = os.listdir(p_dir)
    ap_dir = ''
    for f in things:
        cur_f = os.path.join(p_dir, f)
        if os.path.isdir(cur_f):
            ap_dir = cur_f
            break

    if not ap_dir:
        with open(log_file, 'w') as log:
            log.write(f"Dir not found. {os.listdir(p_dir)}\n")
        cleanup(p_dir)
        return

    # Resolve dependencies
    #if not resolve_dependencies(p_dir):
    #    with open(log_file, 'a') as log:
    #        log.write("Failed to resolve dependencies.\n")
    #    return

    # Configure the package
    stdout, stderr, success, corm = configure_and_build_package(ap_dir)
    if success:
        print('.', end='')
        with lock:
            if corm:
                configure_success += 1
                with open(csuccess_file, 'a') as f:
                    f.write(package + '\n')
            else:
                make_success += 1
                with open(msuccess_file, 'a') as g:
                    g.write(package + '\n')
    else:
        with open(log_file, 'a') as log:
            log.write(stderr)
        cleanup(p_dir)
        return

def thread_routine():
    while True:
        package = queue.get()
        if package is None:
            break
        process_package(package)
        queue.task_done()

# Create threads
threads = []
for i in range(NUM_THREADS):
    t = threading.Thread(target=thread_routine)
    t.start()
    threads.append(t)

# Enqueue packages
for i in range(total_packages):
    queue.put(packages[i])

# Block until all tasks are done
queue.join()

# Stop workers
for i in range(NUM_THREADS):
    queue.put(None)
for t in threads:
    t.join(120)

# Print success rates
print(f"Total packages: {total_packages}")
print(f"Configure success: {configure_success} ({(configure_success/total_packages)*100:.2f}%)")
print(f"Make success: {make_success} ({(make_success/total_packages)*100:.2f}%)")

